@@ -1,11 +1,8 @@
 """ Module for sky subtraction
 
 .. include common links, assuming primary doc root is up one directory
-.. include:: ../links.rst
+.. include:: ../include/links.rst
 """
-import sys
-import os
-
 import numpy as np
 
 from scipy import ndimage
@@ -17,8 +14,9 @@ from IPython import embed
 
 from pypeit.images import imagebitmask
 from pypeit.core import basis, pixels, extract
-from pypeit import msgs, utils, ginga, bspline, slittrace
-from pypeit.core.moment import moment1d
+from pypeit.core import fitting
+from pypeit import msgs, utils, bspline, slittrace
+from pypeit.display import display
 
 def skysub_npoly(thismask):
     """
@@ -94,8 +92,7 @@ def global_skysub(image, ivar, tilts, thismask, slit_left, slit_righ, inmask=Non
             is closed.
 
     Returns:
-        `numpy.ndarray`_: Returns the model sky background at the pixels
-        where thismask is True::
+        `numpy.ndarray`_ : The model sky background at the pixels where thismask is True::
 
             >>>  skyframe = np.zeros_like(image)
             >>>  thismask = slitpix == thisslit
@@ -121,7 +118,7 @@ def global_skysub(image, ivar, tilts, thismask, slit_left, slit_righ, inmask=Non
         msgs.error("Type of inmask should be bool and is of type: {:}".format(inmask.dtype))
 
     # Sky pixels for fitting
-    inmask_in = thismask & (ivar > 0.0) & inmask & np.invert(edgmask)
+    inmask_in = thismask & (ivar > 0.0) & inmask & np.logical_not(edgmask)
     isrt = np.argsort(piximg[thismask])
     pix = piximg[thismask][isrt]
     sky = image[thismask][isrt]
@@ -140,7 +137,7 @@ def global_skysub(image, ivar, tilts, thismask, slit_left, slit_righ, inmask=Non
             #lsky_ivar = np.full(lsky.shape, 0.1)
             # Init bspline to get the sky breakpoints (kludgy)
             lskyset, outmask, lsky_fit, red_chi, exit_status \
-                    = utils.bspline_profile(pix[pos_sky], lsky, lsky_ivar, np.ones_like(lsky),
+                    = fitting.bspline_profile(pix[pos_sky], lsky, lsky_ivar, np.ones_like(lsky),
                                             ingpm=inmask_fit[pos_sky], upper=sigrej, lower=sigrej,
                                             kwargs_bspline={'bkspace':bsp},
                                             kwargs_reject={'groupbadpix': True, 'maxrej': 10})
@@ -160,7 +157,7 @@ def global_skysub(image, ivar, tilts, thismask, slit_left, slit_righ, inmask=Non
     # Perform the full fit now
     msgs.info("Full fit in global sky sub.")
     skyset, outmask, yfit, _, exit_status \
-            = utils.bspline_profile(pix, sky, sky_ivar, poly_basis, ingpm=inmask_fit, nord=4,
+            = fitting.bspline_profile(pix, sky, sky_ivar, poly_basis, ingpm=inmask_fit, nord=4,
                                     upper=sigrej, lower=sigrej, maxiter=maxiter,
                                     kwargs_bspline={'bkspace':bsp},
                                     kwargs_reject={'groupbadpix':True, 'maxrej': 10})
@@ -175,7 +172,7 @@ def global_skysub(image, ivar, tilts, thismask, slit_left, slit_righ, inmask=Non
         poly_basis = np.ones_like(sky)
         # Perform the full fit now
         skyset, outmask, yfit, _, exit_status \
-                = utils.bspline_profile(pix, sky, sky_ivar, poly_basis, ingpm=inmask_fit, nord=4,
+                = fitting.bspline_profile(pix, sky, sky_ivar, poly_basis, ingpm=inmask_fit, nord=4,
                                         upper=sigrej, lower=sigrej, maxiter=maxiter,
                                         kwargs_bspline={'bkspace': bsp},
                                         kwargs_reject={'groupbadpix': False, 'maxrej': 10})
@@ -271,7 +268,7 @@ def skyoptimal(wave, data, ivar, oprof, sortpix, sigrej=3.0, npoly=1, spatial=No
 
     if ngood > 0:
         sset1, outmask_good1, yfit1, red_chi1, exit_status \
-                = utils.bspline_profile(wave[good], data[good], ivar[good], profile_basis[good, :],
+                = fitting.bspline_profile(wave[good], data[good], ivar[good], profile_basis[good, :],
                                         fullbkpt=fullbkpt, upper=sigrej, lower=sigrej,
                                         relative=relative,
                                         kwargs_reject={'groupbadpix': True, 'maxrej': 5})
@@ -290,7 +287,7 @@ def skyoptimal(wave, data, ivar, oprof, sortpix, sigrej=3.0, npoly=1, spatial=No
     msgs.info('Iter     Chi^2     Rejected Pts')
     if np.any(mask1):
         sset, outmask_good, yfit, red_chi, exit_status \
-                = utils.bspline_profile(wave[good], data[good], ivar[good], profile_basis[good,:],
+                = fitting.bspline_profile(wave[good], data[good], ivar[good], profile_basis[good,:],
                                         ingpm=mask1, fullbkpt=fullbkpt, upper=sigrej, lower=sigrej,
                                         relative=relative,
                                         kwargs_reject={'groupbadpix': True, 'maxrej': 1})
@@ -473,7 +470,7 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
                          spat_pix=None, adderr=0.01, bsp=0.6, extract_maskwidth=4.0, trim_edg=(3,3),
                          std=False, prof_nsigma=None, niter=4, box_rad=7, sigrej=3.5, bkpts_optimal=True,
                          debug_bkpts=False,sn_gauss=4.0, model_full_slit=False, model_noise=True, show_profile=False,
-                         show_resids=False, use_2dmodel_mask=True):
+                         show_resids=False, use_2dmodel_mask=True, no_local_sky=False):
     """Perform local sky subtraction and  extraction
 
      Args:
@@ -598,6 +595,9 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
             Show the
         use_2dmodel_mask (bool, optional):
             Use the mask made from profile fitting when extracting?
+        no_local_sky (bool, optional):
+            If True, do not fit local sky model, only object profile and extract optimally
+            The objimage will be all zeros.
 
     Returns:
         :obj:`tuple`:  Returns (skyimage[thismask], objimage[thismask],
@@ -623,9 +623,9 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
     # Set up the prof_nsigma
     if (prof_nsigma is None):
         prof_nsigma1 = np.full(len(sobjs), None)
-    elif len(prof_nsigma) == 1:
+    elif np.size(prof_nsigma) == 1:
         prof_nsigma1 = np.full(nobj, prof_nsigma)
-    elif len(prof_nsigma) == nobj:
+    elif np.size(prof_nsigma) == nobj:
         prof_nsigma1 = prof_nsigma
     else:
         raise ValueError('Invalid size for prof_nsigma.')
@@ -735,7 +735,7 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
                     extract.extract_optimal(sciimg, modelivar, (outmask & objmask), waveimg, skyimage, rn2_img, thismask,
                                             last_profile, box_rad, sobjs[iobj])
                     # If the extraction is bad do not update
-                    if 'OPT_MASK' in sobjs[iobj].keys():
+                    if sobjs[iobj].OPT_MASK is not None:
                         if sobjs[iobj].OPT_MASK.any():
                             flux = sobjs[iobj].OPT_COUNTS
                             fluxivar = sobjs[iobj].OPT_COUNTS_IVAR*sobjs[iobj].OPT_MASK
@@ -751,8 +751,6 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
                         thisfwhm=sobjs[iobj].FWHM, maskwidth=sobjs[iobj].maskwidth,
                         prof_nsigma=sobjs[iobj].prof_nsigma, sn_gauss=sn_gauss, obj_string=obj_string,
                         show_profile=show_profile)
-                    #proc_list.append(show_proc)
-
                     # Update the object profile and the fwhm and mask parameters
                     obj_profiles[ipix[0], ipix[1], ii] = profile_model
                     sobjs[iobj].TRACE_SPAT = trace_new
@@ -760,18 +758,16 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
                     sobjs[iobj].FWHM = np.median(fwhmfit)
                     mask_fact = 1.0 + 0.5 * np.log10(np.fmax(np.sqrt(np.fmax(med_sn2, 0.0)), 1.0))
                     maskwidth = extract_maskwidth*np.median(fwhmfit) * mask_fact
-                    if sobjs[iobj].prof_nsigma is None:
-                        sobjs[iobj].maskwidth = maskwidth
-                    else:
-                        sobjs[iobj].maskwidth = sobjs[iobj].prof_nsigma * (sobjs[iobj].FWHM / 2.3548)
-
+                    sobjs[iobj].maskwidth = maskwidth if sobjs[iobj].prof_nsigma is None else \
+                        sobjs[iobj].prof_nsigma * (sobjs[iobj].FWHM / 2.3548)
                 else:
                     msgs.warn("Bad extracted wavelengths in local_skysub_extract")
                     msgs.warn("Skipping this profile fit and continuing.....")
 
+            # Fit the local sky
             sky_bmodel = np.array(0.0)
             iterbsp = 0
-            while (not sky_bmodel.any()) & (iterbsp <= 4):
+            while (not sky_bmodel.any()) & (iterbsp <= 4) & (not no_local_sky):
                 bsp_now = (1.2 ** iterbsp) * bsp
                 fullbkpt = optimal_bkpts(bkpts_optimal, bsp_now, piximg, localmask, debug=(debug_bkpts & (iiter == niter)),
                                          skyimage=skyimage, min_spat=min_spat, max_spat=max_spat)
@@ -783,9 +779,9 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
 
                 skymask = outmask & np.invert(edgmask)
                 sky_bmodel, obj_bmodel, outmask_opt = skyoptimal(
-                    piximg.flat[isub], sciimg.flat[isub], (modelivar * skymask).flat[isub],
-                    obj_profiles_flat[isub, :], sortpix, spatial=spatial_img.flat[isub],
-                    fullbkpt=fullbkpt, sigrej=sigrej_eff, npoly=npoly)
+                        piximg.flat[isub], sciimg.flat[isub], (modelivar * skymask).flat[isub],
+                        obj_profiles_flat[isub, :], sortpix, spatial=spatial_img.flat[isub],
+                        fullbkpt=fullbkpt, sigrej=sigrej_eff, npoly=npoly)
                 iterbsp = iterbsp + 1
                 if (not sky_bmodel.any()) & (iterbsp <= 3):
                     msgs.warn('***************************************')
@@ -835,10 +831,13 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
                         'Iteration = {:d}'.format(iiter) + ', rejected {:d}'.format(nrej) + ' of ' + '{:d}'.format(
                             igood1.sum()) + 'fit pixels')
 
+            elif no_local_sky:
+                pass
             else:
                 msgs.warn('ERROR: Bspline sky subtraction failed after 4 iterations of bkpt spacing')
                 msgs.warn('       Moving on......')
                 obj_profiles = np.zeros_like(obj_profiles)
+                isub, = np.where(localmask.flatten())
                 # Just replace with the global sky
                 skyimage.flat[isub] = global_sky.flat[isub]
 
@@ -866,7 +865,7 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
 
     # If requested display the model fits for this slit
     if show_resids:
-        viewer, ch = ginga.show_image((sciimg - skyimage - objimage) * np.sqrt(modelivar) * thismask)
+        viewer, ch = display.show_image((sciimg - skyimage - objimage) * np.sqrt(modelivar) * thismask)
         # TODO add error checking here to see if ginga exists
         canvas = viewer.canvas(ch._chname)
         out1 = canvas.clear()
@@ -878,7 +877,7 @@ def local_skysub_extract(sciimg, sciivar, tilts, waveimg, global_sky, rn2_img,
                 color = 'magenta'
             else:
                 color = 'orange'
-            ginga.show_trace(viewer, ch, spec.TRACE_SPAT, spec.name, color=color)
+            display.show_trace(viewer, ch, spec.TRACE_SPAT, spec.NAME, color=color)
 
         # These are the pixels that were masked by the extraction
         spec_mask, spat_mask = np.where((outmask == False) & (inmask == True))
@@ -1034,11 +1033,12 @@ def ech_local_skysub_extract(sciimg, sciivar, fullmask, tilts, waveimg, global_s
                         minx = 0.0
                         maxx = fwhm_here[other_orders].max()
                         # ToDO robust_poly_fit needs to return minv and maxv as outputs for the fits to be usable downstream
-                        fit_mask, fwhm_coeffs = utils.robust_polyfit_djs(order_vec[other_orders], fwhm_here[other_orders],1,
+                        #fit_mask, fwhm_coeffs = fitting.robust_fit(order_vec[other_orders], fwhm_here[other_orders],1,
+                        pypeitFit = fitting.robust_fit(order_vec[other_orders], fwhm_here[other_orders],1,
                                                                         function='polynomial',maxiter=25,lower=2.0, upper=2.0,
                                                                         maxrej=1,sticky=False, minx=minx, maxx=maxx)
-                        fwhm_this_ord = utils.func_val(fwhm_coeffs, order_vec[iord], 'polynomial', minx=minx, maxx=maxx)
-                        fwhm_all = utils.func_val(fwhm_coeffs, order_vec, 'polynomial', minx=minx, maxx=maxx)
+                        fwhm_this_ord = pypeitFit.eval(order_vec[iord])#, 'polynomial', minx=minx, maxx=maxx)
+                        fwhm_all = pypeitFit.eval(order_vec)#, 'polynomial', minx=minx, maxx=maxx)
                         fwhm_str = 'linear fit'
                     else:
                         fit_mask = np.ones_like(order_vec[other_orders],dtype=bool)
@@ -1178,7 +1178,7 @@ def read_userregions(skyreg, nslits, maxslitlength):
     return status, skyreg
 
 
-def generate_mask(pypeline, skyreg, slits, slits_left, slits_right):
+def generate_mask(pypeline, skyreg, slits, slits_left, slits_right, spat_flexure=None):
     """Generate the mask of sky regions
 
     Parameters
@@ -1239,4 +1239,4 @@ def generate_mask(pypeline, skyreg, slits, slits_left, slits_right):
                                      mask=slmsk, specmin=spec_min, specmax=spec_max,
                                      binspec=slits.binspec, binspat=slits.binspat, pad=0)
     # Generate the mask, and return
-    return (slitreg.slit_img(use_spatial=False) >= 0).astype(np.bool)
+    return (slitreg.slit_img(use_spatial=False, flexure=spat_flexure) >= 0).astype(np.bool)
